from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
from tensorflow.python.keras import testing_utils
from .. import WeightNorm


class WeightNormWrapperTest(tf.test.TestCase):
    def testDoubleWrap(self):
        dense_layer = tf.layers.Dense(7)

        WeightNorm(dense_layer)
        with self.assertRaisesRegexp(ValueError, 'Weight normalization already applied'):
            WeightNorm(dense_layer)

    def testNoKernel(self):
        with self.assertRaisesRegexp(ValueError, 'Parameter .* not found in layer'):
            WeightNorm(tf.keras.layers.MaxPooling2D(2, 2)).build((2, 2))

    def testVarsAndShapes(self):
        source_input = tf.random_normal([3, 5])

        dense_layer = tf.layers.Dense(7)
        dense_wrapper = WeightNorm(dense_layer)
        weighted_output = dense_wrapper(source_input)

        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            sess.run(weighted_output)

        self.assertEqual(dense_layer.kernel_v.shape.as_list(), dense_layer.kernel.shape.as_list())
        self.assertEqual(dense_layer.kernel_g.shape, [7])

    def testDenseDecomposition(self):
        source_input = tf.random_normal([3, 5])

        original_layer = tf.layers.Dense(7)
        original_output = original_layer(source_input)

        wrapper_layer = WeightNorm(original_layer)
        weighted_output = wrapper_layer(source_input)

        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            original_result, weighted_result = sess.run([original_output, weighted_output])

        self.assertAllClose(original_result, weighted_result)
        self.assertNotEqual(original_result.tolist(), weighted_result.tolist())

    def testConvDecomposition(self):
        source_input = tf.random_normal([3, 5, 7, 9])

        original_layer = tf.layers.Conv2D(2, 4)
        original_output = original_layer(source_input)

        wrapper_layer = WeightNorm(original_layer)
        weighted_output = wrapper_layer(source_input)

        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            original_result, weighted_result = sess.run([original_output, weighted_output])

        self.assertAllClose(original_result, weighted_result)
        self.assertNotEqual(original_result.tolist(), weighted_result.tolist())

    def testLayer(self):
        with self.cached_session(use_gpu=True):
            with tf.keras.utils.custom_object_scope({'WeightNorm': WeightNorm}):
                testing_utils.layer_test(
                    WeightNorm,
                    kwargs={
                        'layer': tf.layers.Dense(1),
                    },
                    input_shape=(3, 7)
                )


class WeightNormModelTest(tf.test.TestCase):
    def setUp(self):
        self.num_classes = 10

        (train_x, train_y), _ = tf.keras.datasets.cifar10.load_data()
        self.train_x = train_x[:10].astype(np.float32) / 255
        self.train_y = tf.keras.utils.to_categorical(train_y[:10].astype(np.int), self.num_classes)

    def testFunctionStyle(self):
        model = tf.keras.Sequential()
        model.add(WeightNorm(tf.keras.layers.Conv2D(6, 5, activation='relu')))
        model.add(tf.keras.layers.MaxPooling2D(2, 2))
        model.add(tf.keras.layers.Flatten())
        model.add(WeightNorm(tf.keras.layers.Dense(self.num_classes, activation='softmax')))
        model.compile(optimizer=tf.train.AdamOptimizer(), loss='categorical_crossentropy', metrics=['accuracy'])
        model.fit(self.train_x, self.train_y, epochs=1)

    def testClassStyle(self):
        num_classes = self.num_classes

        class TestModel(tf.keras.Model):
            def __init__(self):
                super(TestModel, self).__init__()
                self.conv = WeightNorm(tf.keras.layers.Conv2D(6, 5, activation='relu'))
                self.maxpool = tf.keras.layers.MaxPooling2D(2, 2)
                self.flatten = tf.keras.layers.Flatten()
                self.dense = WeightNorm(tf.keras.layers.Dense(num_classes, activation='softmax'))

            def call(self, input, training=None, mask=None):
                x = self.conv(input)
                x = self.maxpool(x)
                x = self.flatten(x)
                x = self.dense(x)

                return x

        model = TestModel()
        model.compile(optimizer=tf.train.AdamOptimizer(), loss='categorical_crossentropy', metrics=['accuracy'])
        model.fit(self.train_x, self.train_y, epochs=1)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='tfwn',
    version='1.0.1',
    description='Weight normalization layer for TensorFlow',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/shkarupa-alex/tfwn',
    author='Shkarupa Alex',
    author_email='shkarupa.alex@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        # https://github.com/tensorflow/tensorflow/issues/7166
        # 'tensorflow>=1.12.0',
    ],
    extras_require={
        'tf_cpu': ['tensorflow>=1.12.0'],
        'tf_gpu': ['tensorflow-gpu>=1.12.0'],
    },
    test_suite='nose.collector',
    tests_require=[
        'nose',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow.python.keras import backend as K

_DEFAULT_WEIGHT_NAME = 'kernel'


class WeightNorm(tf.keras.layers.Wrapper):
    """ Applies weight normalization to a layer. Weight normalization is a reparameterization that decouples the
    magnitude of a weight tensor from its direction. This speeds up convergence by improving the conditioning of the
    optimization problem.

    Reference: https://arxiv.org/abs/1602.07868
    Weight Normalization: A Simple Reparameterization to Accelerate Training of Deep Neural Networks
    Tim Salimans, Diederik P. Kingma (2016)

    Arguments:
      layer: The `Layer` instance to be wrapped.
      weight_name: name of weights `Tensor` in wrapped layer to be normalized
    """

    def __init__(self, layer, weight_name=_DEFAULT_WEIGHT_NAME, **kwargs):
        self.weight_name = weight_name

        if not hasattr(layer, '_weight_norm'):
            layer._weight_norm = {weight_name}
        elif weight_name not in layer._weight_norm:
            layer._weight_norm.add(weight_name)
        else:
            raise ValueError(
                'Weight normalization already applied to parameter {} in layer {}'.format(weight_name, layer))

        self.g_name = '{}_g'.format(weight_name)
        self.v_name = '{}_v'.format(weight_name)

        super(WeightNorm, self).__init__(layer, **kwargs)
        self.supports_masking = True
        self._track_checkpointable(layer, name='layer')

    def build(self, input_shape=None):
        self.input_spec = tf.keras.layers.InputSpec(shape=tf.TensorShape(input_shape).as_list())

        if not self.layer.built:
            self.layer.build(input_shape)

        if hasattr(self.layer, self.g_name) or hasattr(self.layer, self.v_name):
            raise ValueError(
                'Weight normalization already applied to {} in layer {}'.format(self.weight_name, self.layer))

        if not hasattr(self.layer, self.weight_name):
            raise ValueError('Parameter {} not found in layer {}'.format(self.weight_name, self.layer))

        v = getattr(self.layer, self.weight_name)
        weight_depth = v.shape[-1]
        norm_axes = list(range(v.shape.ndims - 1))

        def g_init(v_init):
            c = K.shape(v_init)[-1]
            v_init = K.reshape(v_init, [-1, c])

            return K.linalg_ops.norm(v_init, ord=2, axis=0)

        self.layer.built = False
        g = self.layer.add_variable(
            name=self.g_name,
            shape=(weight_depth,),
            initializer=lambda *args, **kwargs: g_init(v.initialized_value()),
            dtype=v.dtype,
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN
        )
        self.layer.built = True

        w = K.nn.l2_normalize(v, axis=norm_axes) * g

        setattr(self.layer, self.g_name, g)
        setattr(self.layer, self.v_name, v)
        setattr(self.layer, self.weight_name, w)

        super(WeightNorm, self).build()

    def call(self, inputs, **kwargs):
        return self.layer.call(inputs, **kwargs)

    def compute_output_shape(self, input_shape):
        return self.layer.compute_output_shape(input_shape)

    def get_config(self):
        config = super(WeightNorm, self).get_config()

        if _DEFAULT_WEIGHT_NAME != self.weight_name:
            config['weight_name'] = self.weight_name

        return config
